/*
 * gtkpasman
 * Copyright (C) 2003 Cyril Bellot
 *
 * passwd.c
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <gtk/gtk.h>
#include <regex.h>
#include "config.h"
#include "interface.h"

#define MAX_CMDLINE 1000

/*
 * readpasswdline
 * 	Retrieve values from the password file
 * 	fields separated by ;
 */
int
readpasswdline (gchar * line, gchar * res[])
{
  gchar **fields;
  gchar tmp[512];
  gchar *tmp2;
  gint nb = 0, i;
  gchar *pattern[] = {
    "customer=%[^;]",
    "group=%[^;]",
    "server=%[^;]",
    "service=%[^;]",
    "serviceplus=%[^;]",
    "user=%[^;]",
    "password=%[^\n]"
  };

  /* Split fields */
  fields = g_strsplit (line, ";", 0);
  for (i = 0; i < 7; i++) {
    tmp[0] = tmp[1] = 0;
    /* strip spaces at start+end, and find pattern match */
    if (sscanf (g_strstrip (fields[i]), pattern[i], tmp)) {
      res[i + 1] = g_strdup (tmp);
      nb++;
    }
  }

  /* If passwd contains ; it will be cut. We stick them here */
  for (i = 7; fields[i] != NULL; i++) {
    tmp2 = g_strdup_printf ("%s;%s", res[7], fields[i]);
    g_free (res[7]);
    res[7] = g_strdup (tmp2);
    g_free (tmp2);
  }

  /* Free resources */
  for (i = 0; fields[i] != NULL; i++)
    g_free (fields[i]);

  /* Number of matched fields */
  return nb;
}

/*
 * give_command
 * 	give command to execute depending on service, replacing 
 * 	%u by username
 * 	%h by hostname
 */
gchar *
give_command (char *service, char *host, char *user)
{
  gchar c, *s;
  gchar *cmd, *res, *tmp;

  if (!g_strcasecmp (service, "Ssh")) {
    cmd = g_strdup (sshcmd);
  }
  else if (!g_strcasecmp (service, "Ftp")) {
    cmd = g_strdup (ftpcmd);
  }
  else if (!g_strcasecmp (service, "Telnet")) {
    cmd = g_strdup (telnetcmd);
  }
  else
    return "";

  /* replace %u or %h by values */
  res = g_strdup ("");
  tmp = g_strdup ("");
  for (s = cmd; *s != '\0'; ++s) {
    if (*s == '%' && *(s + 1) != '\0') {
      if ((c = *(s + 1)) == 'h')
	tmp = g_strconcat (res, host, NULL);
      else if (c == 'u') {
	tmp = g_strconcat (res, user, NULL);
      }
      s++;
    }
    else {
      tmp = g_strdup_printf ("%s%c", res, *s);
    }
    g_free (res);
    res = g_strdup (tmp);
    g_free (tmp);
  }

  /* free resources */
  g_free (cmd);

  return res;
}

/*
 * emptytree_data
 * 	callback to free the 3 blocks of data linked to elements of the tree
 */
void
emptytree_data (gpointer data)
{
  g_free (((gchar **) data)[0]);
  g_free (((gchar **) data)[1]);
  g_free (((gchar **) data)[2]);
  g_free (((gchar **) data));
}

/*
 * emptytree
 * 	free clist tree
 */
void
emptytree (GtkWidget * tree)
{
  gtk_clist_freeze (GTK_CLIST (tree));
  gtk_clist_clear (GTK_CLIST (tree));
  gtk_clist_thaw (GTK_CLIST (tree));
  decrypted_passwds = 0;
}

/*
 * filltree
 * 	add elements to the tree
 * 	add entries to the menu
 */
int
filltree (GtkWidget * tree)
{
  gchar *tmpcustomer, *tmpgroupe, *tmpserver, *tmpservice, *tmpserviceplus;
  GtkCTreeNode *customernode = NULL;
  GtkCTreeNode *groupnode = NULL;
  GtkCTreeNode *hostnode = NULL;
  GtkCTreeNode *servicenode = NULL;
  GtkCTreeNode *serviceplusnode = NULL;
  GtkCTreeNode *descnode = NULL;
  char buf[MAX_CMDLINE + 1];
  int i, passpipe[2];
  int correct = 0, new_before_serviceplus = 0;
  gchar *res[8];
  FILE *fp;
  gchar **data;

  /* freeze clist tree */
  gtk_clist_freeze (GTK_CLIST (tree));

  /* register pipe (to provide password) */
  pipe (passpipe);
  write (passpipe[1], globalpasswd, strlen (globalpasswd));
  close (passpipe[1]);

  /* execute gpg reading password from the pipe */
  g_snprintf (buf, sizeof (buf),
	      "gpg --batch --decrypt --passphrase-fd %d %s 2>&-",
	      passpipe[0], passwdfile);
  fp = popen (buf, "r");
  if (fp == NULL) {
    g_print ("Failed to open pipe\n");
    return 0;
  }

  tmpcustomer = g_strdup ("");
  tmpgroupe = g_strdup ("");
  tmpserver = g_strdup ("");
  tmpservice = g_strdup ("");
  tmpserviceplus = g_strdup ("");
  /* for each line of the decrypted file */
  while (fgets (buf, sizeof (buf), fp) != NULL) {
    if (readpasswdline (buf, res) == 7) {
      correct = 1;
      new_before_serviceplus = 0;
      if (strcmp (tmpcustomer, res[1])) {
	/* new customer */
	customernode = gtk_ctree_insert_node (GTK_CTREE (tree),
					      0, 0, &res[1],
					      0, 0, 0, 0, 0, 0, 0);
	g_free (tmpcustomer);
	tmpcustomer = g_strdup (res[1]);
	g_free (tmpgroupe);
	tmpgroupe = g_strdup ("(none)");
	g_free (tmpserver);
	tmpserver = g_strdup ("(none)");
	g_free (tmpservice);
	tmpservice = g_strdup ("(none)");
	g_free (tmpserviceplus);
	tmpserviceplus = g_strdup ("(none)");
	new_before_serviceplus++;
      }
      if (strcmp (tmpgroupe, res[2])) {
	/* new group */
	if (strcmp (res[2], "")) {
	  /* new branch only is not empty */
	  groupnode = gtk_ctree_insert_node (GTK_CTREE (tree),
					     GTK_CTREE_NODE
					     (customernode), 0,
					     &res[2], 0, 0, 0, 0, 0, 0, 0);
	}
	g_free (tmpgroupe);
	tmpgroupe = g_strdup (res[2]);
	g_free (tmpserver);
	tmpserver = g_strdup ("(none)");
	g_free (tmpservice);
	tmpservice = g_strdup ("(none)");
	g_free (tmpserviceplus);
	tmpserviceplus = g_strdup ("(none)");
	new_before_serviceplus++;
      }
      if (strcmp (tmpserver, res[3])) {
	/* new server */
	if (strcmp (tmpgroupe, "")) {
	  /* linked to group if group is not empty */
	  hostnode = gtk_ctree_insert_node (GTK_CTREE (tree),
					    GTK_CTREE_NODE
					    (groupnode), 0, &res[3],
					    0, 0, 0, 0, 0, 0, 0);
	}
	else {
	  /* linked to customer if group is empty */
	  hostnode = gtk_ctree_insert_node (GTK_CTREE (tree),
					    GTK_CTREE_NODE
					    (customernode), 0,
					    &res[3], 0, 0, 0, 0, 0, 0, 0);
	}
	g_free (tmpserver);
	tmpserver = g_strdup (res[3]);
	g_free (tmpservice);
	tmpservice = g_strdup ("(none)");
	g_free (tmpserviceplus);
	tmpserviceplus = g_strdup ("(none)");
	new_before_serviceplus++;
      }
      if (strcmp (tmpservice, res[4])) {
	/* new service */
	servicenode = gtk_ctree_insert_node (GTK_CTREE (tree),
					     GTK_CTREE_NODE (hostnode),
					     0, &res[4], 0, 0, 0, 0, 0, 0, 0);
	g_free (tmpservice);
	tmpservice = g_strdup (res[4]);
	g_free (tmpserviceplus);
	tmpserviceplus = g_strdup ("(none)");
	new_before_serviceplus++;
      }
      if (strcmp (tmpserviceplus, res[5])) {
	/* new serviceplus */
	if (strcmp (res[5], "")) {
	  /* new branch only if serviceplus is not empty */
	  serviceplusnode = gtk_ctree_insert_node (GTK_CTREE (tree),
						   GTK_CTREE_NODE
						   (servicenode), 0,
						   &res[5], 0, 0, 0,
						   0, 0, 0, 0);
	}
	g_free (tmpserviceplus);
	tmpserviceplus = g_strdup (res[5]);
      }
      /* username readable in menus even if empty... */
      if (!strcmp (res[6], "")) {
	g_free (res[6]);
	res[6] = g_strdup ("__none__");
      }
      /* new user */
      if (strcmp (tmpserviceplus, "")) {
	/* linked to serviceplus if not empty */
	descnode = gtk_ctree_insert_node (GTK_CTREE (tree),
					  GTK_CTREE_NODE
					  (serviceplusnode), 0, &res[6],
					  0, 0, 0, 0, 0, 0, 0);
      }
      else {
	/* otherwise linked to service */
	descnode = gtk_ctree_insert_node (GTK_CTREE (tree),
					  GTK_CTREE_NODE (servicenode),
					  0, &res[6], 0, 0, 0, 0, 0, 0, 0);
      }

      /* data (passwd, commands, infos) */
      data = (gchar **) g_malloc (3 * sizeof (gchar *));
      data[0] = g_strdup (res[7]);
      data[1] = give_command ((char *) tmpservice, (char *) tmpserver,
			      (char *) res[6]);
      data[2] = g_strdup_printf ("%s %s %s", (char *) tmpserver,
				 (char *) tmpservice, (char *) res[6]);
      gtk_ctree_node_set_row_data_full (GTK_CTREE (tree), descnode,
					(gpointer) data, emptytree_data);
      add_to_host_menu (tmpcustomer, tmpgroupe, tmpserver, tmpservice,
			tmpserviceplus, res[6], data);

      /* special case: ssh+root */
      if ((!strcmp (res[6], "root")) && (!strcmp (res[4], "ssh"))) {
	data = (gchar **) g_malloc (3 * sizeof (gchar *));
	data[0] = g_strdup (res[7]);
	data[1] = give_command ((char *) tmpservice, (char *) tmpserver,
				(char *) res[6]);
	data[2] = g_strdup_printf ("%s %s %s", (char *) tmpserver,
				   (char *) tmpservice, (char *) res[6]);
	gtk_ctree_node_set_row_data_full (GTK_CTREE (tree), hostnode,
					  (gpointer) data, emptytree_data);
	add_to_host_menu_sshroot (tmpcustomer, tmpgroupe, tmpserver, data);

      }
      /* free resources */
      for (i = 1; i <= 7; i++) {
	g_free (res[i]);
      }
    }
    else {
      g_print ("WARNING: not matched: %s\n", buf);
    }
  }
  /* close pipe */
  pclose (fp);

  /* unfreeze clist tree */
  gtk_clist_thaw (GTK_CLIST (tree));

  return correct;
}
